import logging
from dataclasses import dataclass, asdict
from functools import partial
from json import load, dump
from pathlib import Path
from typing import List

from rich.prompt import Confirm, Prompt
from telethon import TelegramClient
from telethon.errors import FloodWaitError
from telethon.tl.patched import Message

from telegram_bell.exceptions import SpecifiedChannelDoesNotExist

logger = logging.getLogger("rich")


@dataclass
class SubscribedChannel:
    name: str
    search_keywords: List[str]
    last_id: int = 0

    def update_last_id(self, new_last_id: int, channels_file_path: str):
        self.last_id = new_last_id

        with open(channels_file_path, "r+") as file:
            channels = load(file)
            changed_channels = []
            for channel in channels:
                if channel["name"] == self.name:
                    channel["last_id"] = self.last_id
                changed_channels.append(channel)
            file.seek(0)
            dump(changed_channels, file)
            file.truncate()

    @staticmethod
    def update_existing_channels(channels_file_path: str | Path):
        with open(channels_file_path, "r+") as file:
            channels = load(file)
            changed_channels = []
            for channel in channels:
                if not Confirm.ask(
                    f"Do you want to remove the channel '{channel['name']}'?",
                    default=False,
                ):
                    if search_keywords := Prompt.ask(
                        f"Do you want to change the search keywords of the channel '{channel['name']}'? \n"
                        f"If so, enter the search keywords that will replace the old ones. If not, just press enter",
                        default=False,
                    ):
                        channel["search_keywords"] = search_keywords.split()
                    changed_channels.append(channel)

            while Confirm.ask("Do you want to add a new channel?", default=False):
                name = Prompt.ask("Enter the name of the Telegram channel")
                search_keywords = Prompt.ask(
                    f"Enter your search keywords for the channel '{name}'. Press Enter if you wish to receive a notification for each message."
                ).split()
                changed_channels.append(
                    asdict(SubscribedChannel(name, search_keywords))
                )

            file.seek(0)
            dump(changed_channels, file)
            file.truncate()

    @classmethod
    def read_from_json(
        cls, channels_file_path: str | Path
    ) -> List["SubscribedChannel"]:
        with open(channels_file_path, "r") as file:
            channels = load(file)
        return [SubscribedChannel(**channel) for channel in channels]

    @classmethod
    def show_from_json(cls, channels_file_path: str | Path) -> str | None:
        try:
            channels = SubscribedChannel.read_from_json(channels_file_path)
            return "\n".join([str(channel) for channel in channels])
        except FileNotFoundError:
            logger.error(f"Config doesn't exist. Execute 'tbell config' first, please")

    @classmethod
    def create_config_file(cls, channels_file_path: str | Path):
        channels = []
        while Confirm.ask("Do you want to add a channel?", default=True):
            name = Prompt.ask("Enter the name of the Telegram channel")
            search_keywords = Prompt.ask(
                f"Enter your search keywords for the channel '{name}'. Press Enter if you wish to receive a notification for each message."
            ).split()
            channels.append(SubscribedChannel(name, search_keywords))

        with open(channels_file_path, "w") as channels_file:
            channels = [asdict(channel) for channel in channels]
            dump(channels, channels_file)


def keywords_in_message(keywords: List[str], text: str):
    for keyword in keywords:
        if keyword in text.lower():
            return True


async def read_messages_from_channel(
    client: TelegramClient,
    channel: SubscribedChannel,
    channels_file_path: str | Path,
    discard: bool = False,
):
    reverse = channel.last_id != 0

    try:
        messages = await client.get_messages(
            channel.name, min_id=channel.last_id, reverse=reverse
        )
    except (ValueError, FloodWaitError) as e:
        logger.exception(e)
        raise SpecifiedChannelDoesNotExist(
            "Specified channel doesn't exist. Check your config."
        )
    else:
        await handle_messages(channel, channels_file_path, messages, discard)


discard_messages_from_channel_to_date = partial(
    read_messages_from_channel, discard=True
)


def must_handle_this_message(
    discard: bool, channel: SubscribedChannel, message: Message
):
    return (
        not discard
        and not channel.search_keywords
        or (
            channel.search_keywords
            and message.text
            and keywords_in_message(channel.search_keywords, message.text)
        )
    )


async def handle_messages(
    channel: SubscribedChannel,
    channels_file_path: str,
    messages,
    log: bool = False,
    discard: bool = False,
):
    for message in messages:
        if must_handle_this_message(discard, channel, message):
            if log:
                logger.info(message.sender_id, ":", message.text)
            await message.forward_to("me")
        channel.update_last_id(message.id, channels_file_path)



import logging
import struct
from pathlib import Path

import asyncclick as click
from configclasses.configclasses import dump
from configclasses.exceptions import ConfigFilePathDoesNotExist
from rich.prompt import Confirm
from rich.traceback import install
from telethon import TelegramClient

from telegram_bell.config import Config
from telegram_bell.exceptions import BadAPIConfiguration
from telegram_bell.notifier import (
    SubscribedChannel,
    read_messages_from_channel,
    discard_messages_from_channel_to_date,
)

logger = logging.getLogger("rich")

app_path = Path("~/telegram_bell").expanduser()
app_path.mkdir(exist_ok=True)
config_path = app_path / ".env"
channels_file_path = app_path / "subscribed_channels.json"
telegram_session_path = app_path / "session"


async def setup_telegram_session(config: Config):
    client = TelegramClient(str(telegram_session_path), config.api_id, config.api_hash)

    try:
        async with client:
            subscribed_channels = SubscribedChannel.read_from_json(channels_file_path)
            for channel in subscribed_channels:
                logger.info(
                    f"Updating your subscribed channel '{channel}' and discarding messages up to today"
                )
                await discard_messages_from_channel_to_date(
                    client, channel, channels_file_path
                )
    except struct.error:
        raise BadAPIConfiguration("Execute 'tbell config' with another parameters.")


async def setup_config(config_exists: bool):
    if config_exists:
        if Confirm.ask(
            "Telegram API config already exist, do you want to create a new one?",
            default=False,
        ):
            Config.create(config_path)
        if Confirm.ask(
            f"\nYour channels: \n"
            f"{SubscribedChannel.show_from_json(channels_file_path)} \n"
            f"Do you want to modify these channels?",
            default=False,
        ):
            SubscribedChannel.update_existing_channels(channels_file_path)
        if Confirm.ask(
            "Telegram session already exist, do you want to create a new one? Type y if you want to update your subscribed channels.",
            default=False,
        ):
            config = Config.from_path(config_path)
            await setup_telegram_session(config)
    else:
        Config.create(config_path)
        SubscribedChannel.create_config_file(channels_file_path)
        config = Config.from_path(config_path)
        await setup_telegram_session(config)


@click.group(chain=True)
async def cli():
    pass


@click.command(name="run")
async def run_command():
    try:
        config = Config.from_path(config_path)
    except ConfigFilePathDoesNotExist:
        config = Config.create(config_path)
        dump(config, config_path)
    client = TelegramClient(str(telegram_session_path), config.api_id, config.api_hash)

    try:
        async with client:
            subscribed_channels = SubscribedChannel.read_from_json(channels_file_path)
            while True:
                for channel in subscribed_channels:
                    await read_messages_from_channel(
                        client, channel, channels_file_path
                    )
    except struct.error:
        raise BadAPIConfiguration("Connection error. Please, execute 'tbell config'.")


@click.command(name="config")
async def config_command():
    config_exists = Path(config_path).exists()
    if config_exists and not Confirm.ask(
        "Config already exist, do you want to modify?", default=False
    ):
        return

    await setup_config(config_exists)


@click.command(name="show")
async def show_command():
    print("Your channels:")
    channels = SubscribedChannel.show_from_json(channels_file_path)
    print(channels)


install()
cli.add_command(run_command)
cli.add_command(config_command)
cli.add_command(show_command)

if __name__ == "__main__":
    cli(_anyio_backend="asyncio")
